# -*- coding: utf-8 -*-
"""File generated by generate_code() - 
WARNING! All changes made in this file will be lost!
"""

from ..Classes.import_all import *

load_switch = {
    "Data": Data,
    "Data1D": Data1D,
    "DataFreq": DataFreq,
    "DataLinspace": DataLinspace,
    "DataND": DataND,
    "DataPattern": DataPattern,
    "DataTime": DataTime,
    "Norm_affine": Norm_affine,
    "Norm_func": Norm_func,
    "Norm_indices": Norm_indices,
    "Norm_ref": Norm_ref,
    "Norm_vector": Norm_vector,
    "Normalization": Normalization,
    "RequestedAxis": RequestedAxis,
    "VectorField": VectorField,
}



# -*- coding: utf-8 -*-
# File generated according to Generator/ClassesRef/Norm_affine.csv
# WARNING! All changes made in this file will be lost!
"""Method code available at https://github.com/Eomys/SciDataTool/tree/master/SciDataTool/Methods//Norm_affine
"""

from os import linesep
from sys import getsizeof
from ._check import check_var, raise_
from ..Functions.save import save
from ..Functions.copy import copy
from ..Functions.load import load_init_dict
from ..Functions.Load.import_class import import_class
from .Normalization import Normalization

# Import all class method
# Try/catch to remove unnecessary dependencies in unused method
try:
    from ..Methods.Norm_affine.normalize import normalize
except ImportError as error:
    normalize = error


from ._check import InitUnKnowClassError


class Norm_affine(Normalization):
    """Affine function normalization (slope x axis + offset)"""

    VERSION = 1

    # cf Methods.Norm_affine.normalize
    if isinstance(normalize, ImportError):
        normalize = property(
            fget=lambda x: raise_(
                ImportError("Can't use Norm_affine method normalize: " + str(normalize))
            )
        )
    else:
        normalize = normalize
    # save and copy methods are available in all object
    save = save
    copy = copy

    def __init__(self, slope=1, offset=0, unit="SI", init_dict=None, init_str=None):
        """Constructor of the class. Can be use in three ways :
        - __init__ (arg1 = 1, arg3 = 5) every parameters have name and default values
            for SciDataTool type, -1 will call the default constructor
        - __init__ (init_dict = d) d must be a dictionary with property names as keys
        - __init__ (init_str = s) s must be a string
        s is the file path to load

        ndarray or list can be given for Vector and Matrix
        object or dict can be given for SciDataTool Object"""

        if init_str is not None:  # Load from a file
            init_dict = load_init_dict(init_str)[1]
        if init_dict is not None:  # Initialisation by dict
            assert type(init_dict) is dict
            # Overwrite default value with init_dict content
            if "slope" in list(init_dict.keys()):
                slope = init_dict["slope"]
            if "offset" in list(init_dict.keys()):
                offset = init_dict["offset"]
            if "unit" in list(init_dict.keys()):
                unit = init_dict["unit"]
        # Set the properties (value check and convertion are done in setter)
        self.slope = slope
        self.offset = offset
        # Call Normalization init
        super(Norm_affine, self).__init__(unit=unit)
        # The class is frozen (in Normalization init), for now it's impossible to
        # add new properties

    def __str__(self):
        """Convert this object in a readeable string (for print)"""

        Norm_affine_str = ""
        # Get the properties inherited from Normalization
        Norm_affine_str += super(Norm_affine, self).__str__()
        Norm_affine_str += "slope = " + str(self.slope) + linesep
        Norm_affine_str += "offset = " + str(self.offset) + linesep
        return Norm_affine_str

    def __eq__(self, other):
        """Compare two objects (skip parent)"""

        if type(other) != type(self):
            return False

        # Check the properties inherited from Normalization
        if not super(Norm_affine, self).__eq__(other):
            return False
        if other.slope != self.slope:
            return False
        if other.offset != self.offset:
            return False
        return True

    def compare(self, other, name="self", ignore_list=None):
        """Compare two objects and return list of differences"""

        if ignore_list is None:
            ignore_list = list()
        if type(other) != type(self):
            return ["type(" + name + ")"]
        diff_list = list()

        # Check the properties inherited from Normalization
        diff_list.extend(super(Norm_affine, self).compare(other, name=name))
        if other._slope != self._slope:
            diff_list.append(name + ".slope")
        if other._offset != self._offset:
            diff_list.append(name + ".offset")
        # Filter ignore differences
        diff_list = list(filter(lambda x: x not in ignore_list, diff_list))
        return diff_list

    def __sizeof__(self):
        """Return the size in memory of the object (including all subobject)"""

        S = 0  # Full size of the object

        # Get size of the properties inherited from Normalization
        S += super(Norm_affine, self).__sizeof__()
        S += getsizeof(self.slope)
        S += getsizeof(self.offset)
        return S

    def as_dict(self, type_handle_ndarray=0, keep_function=False, **kwargs):
        """
        Convert this object in a json serializable dict (can be use in __init__).
        type_handle_ndarray: int
            How to handle ndarray (0: tolist, 1: copy, 2: nothing)
        keep_function : bool
            True to keep the function object, else return str
        Optional keyword input parameter is for internal use only
        and may prevent json serializability.
        """

        # Get the properties inherited from Normalization
        Norm_affine_dict = super(Norm_affine, self).as_dict(
            type_handle_ndarray=type_handle_ndarray,
            keep_function=keep_function,
            **kwargs
        )
        Norm_affine_dict["slope"] = self.slope
        Norm_affine_dict["offset"] = self.offset
        # The class name is added to the dict for deserialisation purpose
        # Overwrite the mother class name
        Norm_affine_dict["__class__"] = "Norm_affine"
        return Norm_affine_dict

    def _set_None(self):
        """Set all the properties to None (except SciDataTool object)"""

        self.slope = None
        self.offset = None
        # Set to None the properties inherited from Normalization
        super(Norm_affine, self)._set_None()

    def _get_slope(self):
        """getter of slope"""
        return self._slope

    def _set_slope(self, value):
        """setter of slope"""
        check_var("slope", value, "float")
        self._slope = value

    slope = property(
        fget=_get_slope,
        fset=_set_slope,
        doc=u"""slope of the axis function

        :Type: float
        """,
    )

    def _get_offset(self):
        """getter of offset"""
        return self._offset

    def _set_offset(self, value):
        """setter of offset"""
        check_var("offset", value, "float")
        self._offset = value

    offset = property(
        fget=_get_offset,
        fset=_set_offset,
        doc=u"""offset of the axis function

        :Type: float
        """,
    )



# -*- coding: utf-8 -*-
from SciDataTool.Classes.Data import Data
from SciDataTool.Classes.Data1D import Data1D
from SciDataTool.Classes.DataFreq import DataFreq
from SciDataTool.Classes.DataTime import DataTime
from SciDataTool.Classes.DataLinspace import DataLinspace
from SciDataTool.Classes.DataND import DataND
from SciDataTool.Classes.RequestedAxis import RequestedAxis
from SciDataTool.Classes.VectorField import VectorField
from SciDataTool.Classes.DataPattern import DataPattern
from SciDataTool.Classes.Norm_affine import Norm_affine
from SciDataTool.Classes.Norm_func import Norm_func
from SciDataTool.Classes.Norm_vector import Norm_vector
from SciDataTool.Classes.Norm_indices import Norm_indices
from SciDataTool.Classes.Norm_ref import Norm_ref

__version__ = "1.4.10"



# -*- coding: utf-8 -*-
from SciDataTool import Data1D
from SciDataTool.Functions import AxisError, axes_dict, rev_axes_dict
from SciDataTool.Functions.conversions import get_unit_derivate, get_unit_integrate


def get_data_along(self, *args, unit="SI", is_norm=False, axis_data=[]):
    """Returns the sliced or interpolated version of the data, using conversions and symmetries if needed.
    Parameters
    ----------
    self: Data
        a Data object
    *args: list of strings
        List of axes requested by the user, their units and values (optional)
    unit: str
        Unit requested by the user ("SI" by default)
    is_norm: bool
        Boolean indicating if the field must be normalized (False by default)
    axis_data: list
        list of ndarray corresponding to user-input data
    Returns
    -------
    a DataND object
    """

    results = self.get_along(*args, is_squeeze=False)
    values = results.pop(self.symbol)
    del results["axes_dict_other"]
    axes_list = results.pop("axes_list")
    Axes = []

    axes_name_new = list(results.keys())
    if "time" in axes_name_new:
        Data_type = "DataTime"
    elif "freqs" in axes_name_new:
        Data_type = "DataFreq"
    else:
        Data_type = "DataND"

    # Dynamic import to avoid loop
    module = __import__("SciDataTool.Classes." + Data_type, fromlist=[Data_type])
    DataClass = getattr(module, Data_type)

    for axis_name in axes_name_new:
        if not isinstance(results[axis_name], str):
            for i, axis in enumerate(self.axes):
                if axis.name == axis_name:
                    index = i
                    name = axis.name
                    is_components = axis.is_components
                    axis_values = results[axis_name]
                    unit = axis.unit
                elif axis_name in axes_dict:
                    if axes_dict[axis_name][0] == axis.name:
                        index = i
                        name = axis_name
                        is_components = axis.is_components
                        axis_values = results[axis_name]
                        unit = axes_dict[axis_name][2]
                elif axis_name in rev_axes_dict:
                    if rev_axes_dict[axis_name][0] == axis.name:
                        index = i
                        name = axis_name
                        is_components = axis.is_components
                        axis_values = results[axis_name]
                        unit = rev_axes_dict[axis_name][2]
            # Update symmetries
            if "smallestperiod" in args[index]:
                symmetries = self.axes[index].symmetries
            else:
                symmetries = dict()
            Axes.append(
                Data1D(
                    name=name,
                    unit=unit,
                    values=axis_values,
                    is_components=is_components,
                    normalizations=self.axes[index].normalizations,
                    symmetries=symmetries,
                ).to_linspace()
            )
    # Update unit if derivation or integration
    for axis in axes_list:
        if axis.extension == "integrate":
            self.unit = get_unit_integrate(self.unit, axis.corr_unit)
        elif axis.extension == "derivate":
            self.unit = get_unit_derivate(self.unit, axis.corr_unit)

    return DataClass(
        name=self.name,
        unit=self.unit,
        symbol=self.symbol,
        axes=Axes,
        values=values,
        normalizations=self.normalizations,
        is_real=self.is_real,
    )



# -*- coding: utf-8 -*-

from SciDataTool.Classes.Data1D import Data1D

import numpy as np
from numpy import (
    ndarray,
    concatenate,
    identity,
    arange,
    asarray,
    sin,
    pi,
    outer,
    linspace,
    sqrt,
)
from sklearn.linear_model import orthogonal_mp
from scipy.fft import idct
from math import floor


def comp_DST(n: int) -> ndarray:
    """
    Compute the matrix of sinusoids, stacked as the columns of DST, with normed columns
    The first null component is removed because it would
    be redondant with the first component of the DCT's matrix.

    Return:
    DST: ndarray of shape (n,n-1)
    """

    f = 0.5 * arange(n)
    t = linspace(0, 1, n, endpoint=False)

    DST = sin(2 * pi * outer(f, t))

    # Norm the columns
    DST = 2 * (1 / sqrt(2 * n)) * DST

    # Remove the first null component
    DST = DST[:, 1:]

    return DST


def comp_undersampling(K: float, Time: Data1D, seed: int = 42) -> ndarray:
    """
    Compute an undersampled Data1D object with a percentage K of the initial samples

    Parameters
    ----------
    K: 0 <= K <= 1
    Time: Data1D object which is undersample
    seed: integer used to initialize the numpy's random generator

    Returns:
    M: ndarray containing the indices of the observed samples of Time
    Time_under: The undersampled version of the Time Data1D object
    """

    assert K <= 1 and K >= 0, "K = {} is not a percentage".format(K)

    n = len(Time.values)
    m = floor(K * n)

    np.random.seed(seed)
    M = np.random.choice(n, m, replace=False)
    M.sort()
    M = asarray(M)

    time_under = Time.values[M]

    Time_under = Data1D(
        name=Time.name,
        unit=Time.unit,
        values=time_under,
    )

    return M, Time_under


def comp_dictionary(n: int, M: ndarray) -> ndarray:
    """
    Construct the dictionary on which the signal is decomposed

    Parameters
    ----------
    n: length of the time vector
    M: index of the grid corresponding to the observations of the undersampled signal

    Returns
    dictionary: concatenation of the DST and DCT's matrix
    """

    DCT = idct(identity(n), type=2, norm="ortho", axis=0)
    DCT = DCT[M]

    # DST with normed columns
    # The first null component is removed in comp_DST
    DST = comp_DST(n)
    DST = DST[M]

    dictionary = concatenate([DCT, DST], axis=1)

    return dictionary


def comp_undersampled_axe(Time: Data1D, Time_under: Data1D) -> ndarray:
    """
    Compute the ndarray M of indices of the undersampled signal such that:
    Time_under.value = Time.value[M]

    This funtion assumes the elements of Time_under.value are elements of Time.value
    """

    # Extract the ndarray time vectors
    time_under = Time_under.values
    time = Time.values

    # Element wise comparison and selection of indices
    M = np.arange(len(time))
    M = M[np.isin(time, time_under)]

    return M


def omp(
    Y: ndarray,
    M: ndarray,
    n: int,
    n_coefs: int = None,
    precompute: bool = True,
    dictionary=None,
    return_path: bool = False,
) -> ndarray:
    """
    Given Y of shape (len(M),n_targets), recover n_targets signals (of length len(M)) with joint sparsity.
    Each signal - column of Y - is the signal's observation on the support M

    Parameter
    ---------
    Y: ndarray (len(M),n_targets) matrix of the n_targets joint sparse signals.
    M: index of the grid corresponding to the observations of the signals.
    n: length of the grid on which the signal is undersampled.
    n_coefs: passed to n_nonzero_coefs, a parameter of orthogonal_mp. It's the number of atoms
    of the dictionary used to decomposed the signals. If None set to 10% of n.
    precompute: whether to precompute. Improves performance for large Y.

    Returns:
    Y_full: ndarray (n,n_targets) matrix of the recovered signals

    """

    if dictionary is None:
        dictionary_decomp = comp_dictionary(n, M)
    else:
        dictionary_decomp = dictionary[0]
        dictionary_synth = dictionary[1]

    # Compute the sparse decomposition of the signals with the scikit-learn
    # This implementation can only be used on real-valued signals, it uses a Cholesky factorization

    if return_path:
        sparse_decomposition, n_iters = orthogonal_mp(
            X=dictionary_decomp,
            y=Y,
            n_nonzero_coefs=n_coefs,
            precompute=precompute,
            return_path=return_path,
        )
    else:
        sparse_decomposition = orthogonal_mp(
            X=dictionary_decomp, y=Y, n_nonzero_coefs=n_coefs, precompute=precompute
        )

    dictionary_synth = comp_dictionary(n, arange(n))

    Y_full = dictionary_synth @ sparse_decomposition

    if return_path:
        return Y_full, n_iters
    else:
        return Y_full



# -*- coding: utf-8 -*-

"""File generated by generate_code() - 
WARNING! All changes made in this file will be lost!
"""

from ..Classes.Data import Data
from ..Classes.Data1D import Data1D
from ..Classes.DataFreq import DataFreq
from ..Classes.DataLinspace import DataLinspace
from ..Classes.DataND import DataND
from ..Classes.DataPattern import DataPattern
from ..Classes.DataTime import DataTime
from ..Classes.Norm_affine import Norm_affine
from ..Classes.Norm_func import Norm_func
from ..Classes.Norm_indices import Norm_indices
from ..Classes.Norm_ref import Norm_ref
from ..Classes.Norm_vector import Norm_vector
from ..Classes.Normalization import Normalization
from ..Classes.RequestedAxis import RequestedAxis
from ..Classes.VectorField import VectorField



from importlib import import_module
from numpy import array

from SciDataTool.Functions.conversions import convert
from SciDataTool.Functions.symmetries import rebuild_symmetries_axis
from SciDataTool.Functions import AxisError, NormError


def get_values(
    self,
    unit="SI",
    is_oneperiod=False,
    is_antiperiod=False,
    is_smallestperiod=False,
    normalization=None,
    operation=None,
    is_real=True,
):
    """Returns the vector 'axis' taking symmetries into account.
    Parameters
    ----------
    self: Data1D
        a Data1D object
    unit: str
        requested unit
    is_oneperiod: bool
        return values on a single period
    is_antiperiod: bool
        return values on a semi period (only for antiperiodic signals)
    is_smallestperiod: bool
        return values on smallest available period
    normalization: str
        name of normalization to use
    operation: str
        name of the operation (e.g. "freqs_to_time")

    Returns
    -------
    Vector of axis values
    """
    # Dynamic import to avoid loop
    module = __import__("SciDataTool.Classes.Norm_vector", fromlist=["Norm_vector"])
    Norm_vector = getattr(module, "Norm_vector")

    if unit != "SI" and unit != self.unit:
        if unit in self.normalizations and normalization is None:
            normalization = unit
            unit = "SI"

    values = self.values
    norm_vector = None

    # Ignore symmetries if fft axis
    if self.name == "freqs" or self.name == "wavenumber":
        is_smallestperiod = True

    # Rebuild symmetries
    if is_smallestperiod:
        pass
    elif is_antiperiod:
        if "antiperiod" in self.symmetries:
            pass
        else:
            raise AxisError("ERROR: axis has no antiperiodicity")
    elif is_oneperiod:
        if "antiperiod" in self.symmetries:
            nper = self.symmetries["antiperiod"]
            self.symmetries["antiperiod"] = 2
            values = rebuild_symmetries_axis(values, self.symmetries)
            self.symmetries["antiperiod"] = nper
            pass
        elif "period" in self.symmetries:
            pass
        else:
            pass
    else:
        values = rebuild_symmetries_axis(values, self.symmetries)
        # if (
        #     normalization is not None
        #     and normalization in self.normalizations
        #     and isinstance(self.normalizations[normalization], Norm_vector)
        # ):
        #     norm_vector = self.normalizations[normalization].vector.copy()
        #     self.normalizations[normalization].vector = rebuild_symmetries_axis(
        #         norm_vector, self.symmetries
        #     )

    # fft/ifft
    if operation is not None:
        module = import_module("SciDataTool.Functions.conversions")
        func = getattr(module, operation)  # Conversion function
        values = array(func(values, is_real=is_real))

    # Normalization
    if normalization is not None:
        if normalization in self.normalizations:
            # if (
            #     self.normalizations[normalization].unit == "SI"
            #     or self.normalizations[normalization].unit == self.unit
            # ):
            # Axis is int he correct unit for the normalization
            # values = self.normalizations[normalization].normalize(values)
            values /= self.normalizations[normalization]
            # else:
            #     raise NormError("Normalization is not available in this unit")
        else:
            raise NormError("Normalization is not available")

    # Unit conversion
    if unit != "SI" and unit != self.unit:
        values = convert(values, self.unit, unit)

    # if norm_vector is not None:
    #     self.normalizations[normalization].vector = norm_vector

    return values
